
//
// assemble.c
//
// <see corresponding .h file>
// 
//
// $Id: assemble.c,v 1.10 2000/12/30 10:39:46 dru Exp $
//
//
//
//

#include "aqmalloc.h"
#include "runtime.h"
#include "atom.h"
#include "aqendian.h"
#include "assemble.h"
#include "nodes.h"



char errMessage[80];


char *  assembleError(void)
{
    return errMessage;
}


void setError(char * str)
{
    strcpy(errMessage, str);
}



// Convert the assembly source into code

// This function takes a prot. It looks up the given slot
// The slot should have an '_assemble' key. If it does, 
// the routine generates a codeproto for the code.
// It then overwrites the original slot in the proto with
// a passthru to this codeproto.
// Example: Root->_Start->_assemble
// You would call assemble on (Root, "_Start")

Proto assemble (Proto proto, Atom slot, int isBlock)
{
    int        i;
    int        size;
    uint32     j;
    int        offset;
    int        k;
    Proto      tmpObj;
    Proto      codeProto;
    Proto      codeBlob;
    Proto      source;
    Proto      assembly;
    Proto      opCode;
    Proto      opType;
    Proto      operand;
    Proto      tmp2;
    uchar *    p;
    int        count;
    
    

    if (isBlock) {
	codeProto = proto;
	// Set the parent, but note that the parent really gets set
	// during the push
	objectSetSlot(codeProto, stringToAtom("parent"), PROTO_ACTION_GET, (uint32) nil );
	objectSetSlot(codeProto, stringToAtom("blockParent"), PROTO_ACTION_GET, (uint32) nil );
    } else {
	// Get the proto that is the codeProto
	objectGetSlot(proto, slot, &codeProto);
	
	// Set the parent
	objectSetSlot(codeProto, stringToAtom("parent"), PROTO_ACTION_GETFRAME, (uint32) objectNewInteger(3) );
    }

    // The source is in the slot "Assemble"
    objectGetSlot(codeProto, stringToAtom("_assemble"), &source);
    
    // Assemble the blob
    
    // Go one pass thru and determine the number of bytes needed
    count = nodeCount(source);
    
    size = 0;
    for (i = 0; i <= count; i++) {
		
	nodeGet(source, i, &assembly);
	
	// Get the data and the op code

	objectGetSlot(assembly, stringToAtom("op"), &opCode);

	if (opCode == (Proto) stringToAtom("slotName")) {
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("nop")) {
	    size++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("break")) {
	    size++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("pop")) {
	    size++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("swap")) {
	    size++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("setupFrame")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("restoreFrame")) {
	    size++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("pushSelf")) {
	    size++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("return")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("pushimmediate")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("argCountCheck")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("call")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("branch")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("branchIfTrue")) {
	    size += 5;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("branchIfFalse")) {
	    size += 5;
	    continue;
	}

	printString("Bad Opcode ");
	printString(atomToString((Atom)opCode));
	printString(" \n");


	exit(1);
    }
    
    
    // Build a Blob
    p = (uchar *) aqcalloc(size);
    codeBlob = objectNewBlob(p);


    // Now assemble the codes into the blob
    i = 0;

    for (i = 0; i <= count; i++) {
	
	nodeGet(source, i, &assembly);
    
	objectGetSlot(assembly, stringToAtom("op"), &opCode);

	if (opCode == (Proto) stringToAtom("slotName")) {
	    objectGetSlot(assembly, stringToAtom("operand"), &operand);
	    objectSetSlot(codeProto, stringToAtom("_Name_"), PROTO_ACTION_GET, (uint32) operand);
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("nop")) {
	    *p = 0x01;
	    p++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("break")) {
	    *p = 0x00;
	    p++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("pop")) {
	    *p = 0x06;
	    p++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("swap")) {
	    *p = 0x07;
	    p++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("restoreFrame")) {
	    *p = 0x09;
	    p++;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("pushSelf")) {
	    *p = 0x0b;
	    p++;
            continue;
	}
	else if (opCode == (Proto) stringToAtom("return")) {
	    *p = 0x05;
	    p++;

	    // simpler integer operand
	    objectGetSlot(assembly, stringToAtom("operand"), &operand);
	    j = objectIntegerValue(operand);

	    *((uint32 *)p) = host2stdInteger(j);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("setupFrame")) {
	    *p = 0x08;
	    p++;

	    // simple integer operand
	    objectGetSlot(assembly, stringToAtom("operand"), &operand);
	    j = objectIntegerValue(operand);

	    *((uint32 *)p) = host2stdInteger(j);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("pushimmediate")) {
	    *p = 0x04;
	    p++;

	    // types should be activateSlot, self, string, or integer
	    objectGetSlot(assembly, stringToAtom("type"), &opType);
	    
	    if (opType == (Proto) stringToAtom("object")) {
		objectGetSlot(assembly, stringToAtom("operand"), &tmpObj );
		j = (uint32) tmpObj;
		if (PROTO_REFTYPE(tmpObj) == PROTO_REF_PROTO) {
		    k = objectGetSlot(tmpObj, stringToAtom("_assemble"), &tmp2);
		    if (k) {
			// Build the block and return a BLOCK reference
			j = (uint32) assemble(tmpObj, (Atom) nil, 1);
		    }
		}
	    }
	    

	    *((uint32 *)p) = host2stdInteger(j);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("argCountCheck")) {
	    *p = 0x0c;
	    p++;

	    // types should be activateSlot, self, string, or integer
	    objectGetSlot(assembly, stringToAtom("type"), &opType);
	    if (opType == (Proto) stringToAtom("self")) {
		// Invalid
                assert(0);
	    }
	    else if (opType == (Proto) stringToAtom("activateSlot")) {
		// Invalid
                assert(0);
	    }
	    else if (opType == (Proto) stringToAtom("object")) {
		// Invalid
		objectGetSlot(assembly, stringToAtom("operand"), &tmpObj );
		j = objectIntegerValue(tmpObj);
	    }
	    
	    *((uint32 *)p) = host2stdInteger(j);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("call")) {
	    *p = 0x02;
	    p++;

	    // types should be activateSlot, self, string, or integer
	    objectGetSlot(assembly, stringToAtom("type"), &opType);
	    if (opType == (Proto) stringToAtom("activateSlot")) {
		j = (uint32) activateSlot;
	    }
	    else if (opType == (Proto) stringToAtom("object")) {
                assert(0);
		objectGetSlot(assembly, stringToAtom("operand"), &tmpObj);
		j = (uint32) tmpObj;
	    }
	    

	    *((uint32 *)p) = host2stdInteger(j);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("branch")) {
	    *p = 0x0d;
	    p++;

	    // simple integer operand
	    objectGetSlot(assembly, stringToAtom("operand"), &operand);
	    offset = objectIntegerValue(operand);

	    *((uint32 *)p) = host2stdInteger(offset);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("branchIfTrue")) {
	    *p = 0x0e;
	    p++;

	    // simple integer operand
	    objectGetSlot(assembly, stringToAtom("operand"), &operand);
	    offset = objectIntegerValue(operand);

	    *((uint32 *)p) = host2stdInteger(offset);
	    
	    p += 4;
	    continue;
	}
	else if (opCode == (Proto) stringToAtom("branchIfFalse")) {
	    *p = 0x0f;
	    p++;

	    // simple integer operand
	    objectGetSlot(assembly, stringToAtom("operand"), &operand);
	    offset = objectIntegerValue(operand);

	    *((uint32 *)p) = host2stdInteger(offset);
	    
	    p += 4;
	    continue;
	}

	exit(1);
    }
    

    // Set the '_code_' slot to be the code Blob
    objectSetSlot(codeProto, stringToAtom("_code_"), PROTO_ACTION_GET, (uint32) codeBlob);

    // Set the '_Assembly' slot to be the code Blob
    //DRUDRU - Should I rename the proto?
    ////objectSetSlot(codeProto, stringToAtom("_assembly"), PROTO_ACTION_GET, (uint32) source);
    
    if (isBlock) {
	codeProto = objectSetAsBlockProto(codeProto);
	return codeProto;
    } else {
	// Set the proto slot to be a passthru to the 
	// code proto
	codeProto = objectSetAsCodeProto(codeProto);
	objectSetSlot(proto, slot, PROTO_ACTION_PASSTHRU, (uint32) codeProto);
    }

    return nil;
}


// Assemble all
// This walks a soup and 
Proto assembleAll(Proto proto)
{
    int i,j;
    struct _protoEntry * slotPtr;
    Proto p1;
    Proto tmp, tmp2;



    // Determine the type of reference this is
    i = (uint32) proto & PROTO_REF_TYPE;
    p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    if (proto == nil) {
	return proto;
    }
    

    //// In theory, a switch doesn't need to be here.
    //// we should put in asserts()
    switch (i) {
	case PROTO_REF_INT:
	    break;
	case PROTO_REF_PROTO:
	    // If this is a forward
	    if (p1->header & PROTO_FLAG_FORWARDER) {
		assembleAll((Proto)p1->count);
	    } else {
		// enumerate through the proto showing it's contents
		// NOTE - since assemble affects the proto that is
		// pointed to by the slot in this proto... we can
		// manipulate it without the worry that we will affect
		// the enumerating process
		j = 0;
		while (1) {
		    slotPtr = objectEnumerate(proto, &j);
		    if (slotPtr == nil) {
			break;	// The end of the enumerate
		    }

		    tmp = protoSlotData(proto, slotPtr);
		    i = (uint32) tmp & PROTO_REF_TYPE;
	
		    if ((PROTO_REF_PROTO == i) && (nil != tmp)) {
			i = objectGetSlot(tmp, stringToAtom("_assemble"), &tmp2);
			if (i) {
			    assemble(proto, (Atom) slotKey(slotPtr), 0);
			} else {
			    assembleAll(tmp);
			}
		    }
		    
		}
	    }
	    
	    
	    break;
	case PROTO_REF_BLOB:
	    break;
	case PROTO_REF_ATOM:
	    break;
    }

    return proto;
    
}
