
//
// bubl2sd.c
//
// <see corresponding .h file>
// 
//
// $Id: bubl2sd.c,v 1.5 2000/12/30 10:39:46 dru Exp $
//
//


#include "crc.h"
#include "bubl.h"
#include "runtime.h"
#include "memstream.h"
#include "bubl2sd.h"


static void * realMapObjects(Proto proto, htab * ht);
static void * realBubl2sd(Proto proto, MemStream ms, htab * ht);
static int    writeSDString(MemStream ms, Proto proto, int32 id); 
static void   assignID(Proto proto, htab * ht);
static int    getNewID(htab * ht);


//
// Brief intro:
// The objects get mapped to a hash table for references (realMapObjects)
// Then the objects get serialized with those ID's (realBubl2sd)
//


// Convert the real objects to SD memory
// 
void *  bubl2sd(Proto proto, unsigned int * length)
{
    MemStream ms;
    uint32    ck;
    char    * str = "AqUa SimpleData\n002\nc  \n";
    htab    * ht;
    
    
    // Setup a memory stream to work over the memory
    ms = NewMemStream((void *) 0, 0);

    // Write header
    writeMSCSizeFrom(ms, strlen(str), str);
    writeMSCInt(ms, 0);		// Length of module - we'll go back and rewrite this

    // Generate hash table of all objects, generate an object->integer mapping
    ht = hcreate(8); 
    // Address zero will hold the last used number - we'll start at 32
    haddInt(ht, 0, (void *)32);
    realMapObjects(proto, ht);

    realBubl2sd(proto, ms, ht);
    
    
    // Write CRC
    *length = getMSSize(ms);
    seekMS(ms, 24);
    writeMSCInt(ms, *length);
    
    seekMS(ms, *length);
    ck = crc32(ms->start, *length);
    writeMSCInt(ms, ck);
    
    *length = getMSSize(ms);
    return ms->start;
}

// This routine will map objects onto a hash table to prepare for
// object serialization
void * realMapObjects(Proto proto, htab * ht)
{
    int i,j;
    struct _protoEntry * slotPtr;
    Proto p1;
    int size;
    Proto tmp;

    // Determine the type of reference this is
    i = (uint32) proto & PROTO_REF_TYPE;
    p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    if (proto == nil) {
	return proto;
    }
    

    switch (i) {
	case PROTO_REF_INT:
	    return nil;
	    break;
	case PROTO_REF_PROTO:
	    // If this is a forward
	    if (p1->header & PROTO_FLAG_FORWARDER) {
		realMapObjects((Proto)p1->count, ht);
	    } else {

	        // Assign the dictionary an id
	        assignID(proto, ht);
		
		i = objectSlotCount(proto);
		
		// enumerate through the proto showing it's contents
		// *Note* we iterate through the slots, not the data
		// section. (Some slots may point to the same data slot)
		j = 0;
		while (1) {
		    slotPtr = objectEnumerate(proto, &j);
		    if (slotPtr == nil) {
			break;
		    }
                    //// Note, this ignores dictionaries that have slots other than 'get'
		    ///  type actions
		    
		    // Map the key
		    realMapObjects((Proto) slotKey(slotPtr), ht);

		    // Map the data
		    realMapObjects(protoSlotData(proto, slotPtr), ht);
		}
	    }
	    
	    break;
	case PROTO_REF_BLOB:
	    // Assign the dictionary an id
	    assignID(proto, ht);
	    
	    //// Need to see if this is a GC'able blob

	    size = objectBlobSize(proto);
	    size = size >> 2;		// Each 4 bytes represents an object
	    
	    
	    // enumerate through the blob
	    for (i = 0; i < size; i++) {
		tmp = objectBlobAtGet(proto, i);
		realMapObjects(tmp, ht);
	    }
	    break;
	case PROTO_REF_ATOM:

	    // Assign the dictionary an id
	    assignID(proto, ht);
	    
	    return proto;
	    break;
    }

    return proto;
}

void assignID(Proto proto, htab * ht)
{
  int i;

  // If it isn't already there, generate a new number
  if ( hfindInt(ht, (ub4) proto) ) {
    // Do nothing
  } else {
    i = getNewID(ht);
    haddInt(ht, (ub4) proto, (void *)i);
  }
}

int getNewID(htab * ht)
{
  int i;

  // Get the counter - no object should be '0'
  if ( hfindInt(ht, 0) ) {
      i = (uint32) hstuff(ht);
      i++;
      // Store it back
      hstuff(ht) = (void *) i;
  } else {
      i = 32;
      haddInt(ht, 0, (void *) i); 
  }
  
  return i;
}


// The routine that does the work on the dumping of objects to the SimpleData format
void * realBubl2sd(Proto proto, MemStream ms, htab * ht)
{
    int32 i;
    int   j;
    Proto p1;
    Proto p;
    struct _protoEntry * slotPtr;
    uint32 size;
    

    // Determine the type of reference this is
    i = (uint32) proto & PROTO_REF_TYPE;
    p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    if (proto == nil) {
	writeMSCChar(ms, 'N');
	return proto;
    }
    

    switch (i) {
	case PROTO_REF_INT:
	    writeMSCChar(ms, 'I');
	    writeMSCInt(ms, objectIntegerValue(proto) );
	    break;
	case PROTO_REF_PROTO:
	    // If this is a forward
	    if (p1->header & PROTO_FLAG_FORWARDER) {
		realBubl2sd((Proto)p1->count, ms, ht);
	    } else {
		
	        assert ( hfindInt(ht, (ub4) proto) );
  	        i = (int32) hstuff(ht);
	        if (i > 0) {
	          // Store the number as a negative value
	          (int32) hstuff(ht) = -i;

     	  	  writeMSCChar(ms, '{'); // Start of dictionary
	          // Write the reference ID
	          writeMSCInt(ms, (int32) i);

		  // Write the data count
		  i = objectDataCount(proto);
	    	  writeMSCChar(ms, 'I');
		  writeMSCInt(ms, i);

		  // Write the slot count
		  j = objectSlotCount(proto);
	    	  writeMSCChar(ms, 'I');
		  writeMSCInt(ms, j);
		
		  // enumerate through the data section of proto
		  for (j = 0; j < i; j++) {
		    realBubl2sd(protoDataAtIndex(proto, j), ms, ht);
		  }

		  // enumerate through the proto showing it's contents
		  // *Note* we iterate through the slots, not the data
		  // section. (Some slots may point to the same data slot)
		  j = 0;
		  while (1) {
		    slotPtr = objectEnumerate(proto, &j);
		    if (slotPtr == nil) {
			break;
		    }
	            // Write the key
		    realBubl2sd((Proto) slotKey(slotPtr), ms, ht);
		    writeMSCChar(ms, slotAction(slotPtr));
		    writeMSCChar(ms, slotIndex(slotPtr));
		  }
		  writeMSCChar(ms, '}' ); // End of dictionary
		} else {
	          // Write the reference
	          writeMSCChar(ms, 'r');
	          writeMSCInt(ms, (int32) -i);
		}
	    }
	    
	    break;
	case PROTO_REF_ATOM:
	    assert ( hfindInt(ht, (ub4) proto) );
	    i = (int32) hstuff(ht);
	    if (i > 0) {
	      // Store the number as a negative value
	      (int32) hstuff(ht) = -i;
	      // Write the reference
	      writeSDString(ms, proto, i);
	    } else {
	      // Write the reference
	      writeMSCChar(ms, 'r');
	      // The integer is already negative, therfore negating it should be
	      // the proper id
	      writeMSCInt(ms, (int32) -i);
	    }
	    break;
	case PROTO_REF_BLOB:
	    // If the blob is internally GC'able, then
	    // Write out as array, else write out as binary
	    // data

	    // NOT IMPLEMENTED YET, so we'll write everything as
	    // an array

	    assert ( hfindInt(ht, (ub4) proto) );
	    i = (int32) hstuff(ht);
	    if (i > 0) {
	      // Store the number as a negative value
	      (int32) hstuff(ht) = -i;
	      // Write the reference

	      size = objectBlobSize(proto);
	      size = size >> 2;
	      
	      // Write string
	      writeMSCChar(ms, '(');
	      // Write reference
	      writeMSCInt(ms, i);
	      writeMSCInt(ms, size);
	      for (i = 0; i < size; i++) {
		  p = objectBlobAtGet(proto, i);
		  realBubl2sd(p, ms, ht);
	      }
	      writeMSCChar(ms, ')');
	    } else {
	      // Write the reference
	      writeMSCChar(ms, 'r');
	      // The integer is already negative, therfore negating it should be
	      // the proper id
	      writeMSCInt(ms, (int32) -i);
	    }
	    break;
        default:
            assert(0);
    }

    return proto;
    
}


int writeSDString(MemStream ms, Proto proto, int32 id) 
{
    char * p;
    int i;
    
    p = atomToString((Atom) proto);
    i = strlen(p);
    i |= 0x80;
    // Write string
    writeMSCChar(ms, (unsigned char) i);
    // Write reference
    writeMSCInt(ms, id);
    writeMSCSizeFrom(ms, (i & 0x7f), p);
    writeMSCChar(ms, '~');
    return 0;
}

