
//
// codegen.c
//
// <see corresponding .h file>
// 
//
// $Id: codegen.c,v 1.5 2000/12/30 10:39:46 dru Exp $
//
//
//
//

#include "bubl.h"
#include "runtime.h"
#include "nodes.h"
#include <ctype.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

// Globals

char stringBuffer[256];


// Function Prototypes

Proto transform(Proto parseTree);

Proto simplifyParseTree(Proto parseTree);

//Proto simplifyExpression(Proto parseTree);

Proto simplifySlots(Proto parseTree);

Proto cleanCodeBlocks(Proto parseTree);

Proto setupBinarySlots(Proto parseTree);

Proto setupKeywordSlots (Proto parseTree);

Proto setupArgSlots(Proto parseTree);

Proto setupMethods(Proto parseTree);

Proto setupBlocks(Proto parseTree);

void countArgsLocals(Proto obj, int *, int *);


// The Big Function
Proto realCodeGen(Proto parseTree);

// Higher Granularity Codegen functions
Proto cgBuildObject(Proto parseTree);

//void  cgBuildSlotList(Proto parseTree, Proto objectUnderConstruction);

void  cgBuildSlot(Proto parseTree, Proto newObj);

Proto cgBuildReceiver(Proto parseTree);

Proto cgBuildReceiverAssembly (Proto parseTree);

Proto cgBuildExpression (Proto parseTree);

Proto cgBuildCodeBlock(Proto parseTree, Proto annotation);

Proto cgBuildMessage(Proto parseTree);

Proto cgBuildUnary(Proto proto);

Proto cgBuildBinary(Proto proto);



//
//  transform - do semantic checks, and simplifications of tree to get it ready for
//  the code generator
// 

Proto transform (Proto parseTree)
{
    Proto newTree;
    
    // Walk the tree
    // and then simplify all nodes that are 0 length

    parseTree = removeForwards(parseTree);

    if (debugParse && (debugParseLevel == 0) ) {
	printString("***No Transform ***\n");
	objectDump(parseTree);
    }

    newTree = simplifyParseTree(parseTree);
    if (debugParse && (debugParseLevel == 1 || debugParseLevel == 0)) {
	printString("*** SimplifyParse ***\n");
	objectDump(newTree);
    }

    newTree = simplifySlots(parseTree);
    if (debugParse && (debugParseLevel == 2 || debugParseLevel == 0)) {
	printString("*** SimplifySlots ***\n");
	objectDump(newTree);
    }

    newTree = cleanCodeBlocks(newTree);
    if (debugParse && (debugParseLevel == 3 || debugParseLevel == 0)) {
	printString("*** CleanCodeBlocks ***\n");
	objectDump(newTree);
    }

    newTree = setupBinarySlots(newTree);
    if (debugParse && (debugParseLevel == 4 || debugParseLevel == 0)) {
	printString("*** MethodBinarySlot ***\n");
	objectDump(newTree);
    }
    newTree = removeForwards(newTree);    

    newTree = setupKeywordSlots(newTree);
    if (debugParse && (debugParseLevel == 5 || debugParseLevel == 0)) {
	printString("*** MethodKeywordSlot ***\n");
	objectDump(newTree);
    }
    newTree = removeForwards(newTree);    

    newTree = setupArgSlots(newTree);
    if (debugParse && (debugParseLevel == 6 || debugParseLevel == 0)) {
	printString("*** setupArgSlot ***\n");
	objectDump(newTree);
    }
    newTree = removeForwards(newTree);    

    newTree = setupMethods(newTree);
    if (debugParse && (debugParseLevel == 7 || debugParseLevel == 0)) {
	printString("*** MethodAnnotate ***\n");
	objectDump(newTree);
    }
    newTree = removeForwards(newTree);    

    newTree = setupBlocks(newTree);
    if (debugParse && (debugParseLevel == 8 || debugParseLevel == 0)) {
 	printString("*** BlockAnnotate ***\n");
 	objectDump(newTree);
    }
    newTree = removeForwards(newTree);    
    
    return newTree;
    
}



//
//  simplify - eliminate the fluff of the parsetree, nodes with no values
// 

Proto simplifyParseTree(Proto parseTree)
{

    int i;
    int size;
    Proto tmp;
    Proto tmpObj;
    Proto proto = parseTree;
    
    // In general, right after a parse most expressions have 
    // lots of nodes and subnodes that are empty. (Because
    // the parse step will match from the topmost production
    // down through the productions to the terminals.)
    // So, we need to simplify the parse tree so the codegen
    // is a lot easier.


    // Walk the parse tree depth first
    // If we encounter Message, Binary, or Unary nodes that have a zero length,
    // then we remove them


    // ACTUAL CODE BELOW HERE

    // Go down each proto, and through each slot.
    // if it hits an expression, go into expression compress mode
    
    // Determine the type of reference this is
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    switch (i) {
	case PROTO_REF_PROTO:
	    if (PROTO_ISFORWARDER(proto)) {
	        assert(0);
	    }
	    
	    size = nodeCount(proto);
	    
	    // enumerate through the proto showing it's contents
	    for (i = 0; i <= size; i++) {
		// Access the data and set it to the new value
		nodeGet(proto, i, &tmp);
		tmp = simplifyParseTree(tmp);
		nodeSet(proto, i, tmp);
	    }
	    

            // Expressions always only have 1 child node
	    objectGetSlot(proto, stringToAtom("name"), &tmp);

	    if ( (tmp == (Proto) stringToAtom("Message")) && (0 == size) ) {
		nodeGet(proto, 0, &tmpObj);
		return tmpObj;
	    }

	    if ( (tmp == (Proto) stringToAtom("Binary")) && (0 == size) ) {
		nodeGet(proto, 0, &tmpObj);
		return tmpObj;
	    }

	    if ( (tmp == (Proto) stringToAtom("Unary")) && (0 == size) ) {
		nodeGet(proto, 0, &tmpObj);
		return tmpObj;
	    }

	    return proto;

	    break;
    }

    // Make the compiler happy
    return proto;
    
}


//
//  simplifySlots   - Remove the zero length expressions from slots
// 

Proto simplifySlots (Proto parseTree)
{

    int i,j;
    int isSimpleSlot = 0;
    Proto p1;
    Proto tmp;
    Proto tmp2;
    Proto proto = parseTree;
    

    // Determine the type of reference this is
    p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }

    
    j = nodeCount(proto);

    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if (tmp == (Proto) stringToAtom("Slot")) {
	nodeGet(proto, 0, &tmp);
// 	if ( (Proto) stringToAtom("Identifier-Slot") == tmp) {
// 	    isSimpleSlot = 0;
// 	}
	if ( (Proto) stringToAtom("Data-Get-Slot") == tmp) {
	    isSimpleSlot = 1;
	}
	if ( (Proto) stringToAtom("Data-SetGet-Slot") == tmp) {
	    isSimpleSlot = 1;
	}
    }

    // enumerate through the proto

    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	simplifySlots(tmp);
    }
    
    if (isSimpleSlot) {
	// Get the contained receiver 
	nodeGet(proto, 2, &tmp);

	// Now get the receiver (notice I'm reusing 'tmp'
	nodeGet(tmp, 0, &tmp2);

 	// Now set the receiver
	nodeSet(proto, 2, tmp2);
    }

    // Make the compiler happy
    return proto;
}





//
//  cleanCodeBlocks - eliminate the codeBlocks with just the "Nill Expression"
//                    statements. Also, remove the 'Nil Expression' nodes
//                    from valid codeBlocks.
// 

Proto cleanCodeBlocks (Proto parseTree)
{

    int i,j;
    int nodeWasNil = 0;
    //Proto p1;
    Proto codeBlock;
    Proto tmp;
    Proto tmpObj;
    Proto tmpName;
    Proto proto = parseTree;

    
    // Go down each proto, and through each slot.
    // if it is an object, it should check for a codeBlock
    // if the codeblock just has a 'Nil Expression' then
    // it should just remove it. Otherwise, it should
    // remove the last 'Nil Expression'
    
    //p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }


    // Enumerate through the proto, recursively 
    // calling this routine

    j = nodeCount(proto);

    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	cleanCodeBlocks(tmp);
    }

    // On the way up, after the routine has finished
    // with the nodes at the bottom of the tree (leaves)
    // We do the real work
    assert(objectGetSlot(proto, stringToAtom("name"), &tmp));
    
    if (tmp == (Proto) stringToAtom("Object")) {
	j = 0;
	assert(nodeGet(proto, 0, &codeBlock));
	// If the '0' slot isn't the 'CodeBlock', then the '1' slot
	// is...
	assert(objectGetSlot(codeBlock, stringToAtom("name"), &tmp));
	if (tmp == (Proto) stringToAtom("SlotList")) {
	    assert(nodeGet(proto, 1, &codeBlock));
	    j = 1;
	}

	// Get the last entry in the code block
	i = nodeCount(codeBlock);
	
	// All this does is verify that the last entry/expression is 
	// a receiver of a "Nil Expression"
	nodeGet(codeBlock, i, &tmpObj);
	assert(objectGetSlot(tmpObj, stringToAtom("name"), &tmpName));
	if (tmpName == (Proto) stringToAtom("Expression")) {
	    assert(objectGetSlot(tmpObj, stringToAtom("_count"), &tmp));
	    if (0 == objectIntegerValue(tmp)) {
		assert(nodeGet(tmpObj, 0, &tmp));
		assert(objectGetSlot(tmp, stringToAtom("name"), &tmpName));
		if (tmpName == (Proto) stringToAtom("Receiver")) {
		    assert(nodeGet(tmp, 0, &tmpName));
		    if (tmpName == (Proto) stringToAtom("Nil Expression")) {
			removeFromNode(codeBlock, i);
			// assert(objectDeleteSlot(codeBlock, stringToAtom(buff)));
			nodeWasNil = 1;
		    }
		}
	    }
	}
	
	// If that was the only expression in the codeBlock, then the 
	// entire codeBlock should be removed.
	if ((0 == i) && (nodeWasNil)) {
	    removeFromNode(proto, j);
	}
    }
    

    // Make the compiler happy
    return proto;
}

//
//  setupBinarySlots - Make the parameters to a binary slot actual members in a slot
//                     list for the object.
// 

Proto setupBinarySlots (Proto parseTree)
{

    int i,j;
    int isObjectBinarySlot = 0;
    Proto p1;
    Proto tmp;
    Proto tmp2;
    Proto slotName;
    Proto argName;
    Proto privateTree;
    Proto newNode;
    Proto proto = parseTree;
    Proto codeObj;
    

    // Determine the type of reference this is
    p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }

    j = nodeCount(proto);

    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if (tmp == (Proto) stringToAtom("Slot")) {
	nodeGet(proto, 0, &tmp);
	if (tmp == (Proto) stringToAtom("Binary-Slot")) {
	    isObjectBinarySlot = 1;
	}
    }

    // enumerate through the proto
    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	setupBinarySlots(tmp);
    }
    
    if (isObjectBinarySlot) {
	// Get the 'Binary-Slot'
	nodeGet(proto, 1, &tmp);
	
	// Get the slotName
	nodeGet(tmp, 0, &slotName);
	
	// Get the argName
	nodeGet(tmp, 1, &argName);
	
	// Get the 'Object'
	nodeGet(tmp, 3, &codeObj);

	newNode = createNode( (Proto) stringToAtom("Slot"));
	addToNode(newNode, (Proto) stringToAtom("Arg-Slot"));
	addToNode(newNode, argName);
	
	// Also add the position in the stack frame for the argument
	addToNode(newNode, (Proto) stringToAtom("8"));

	// See if there is already a slot list
	nodeGet(codeObj, 0, &tmp);
	objectGetSlot(tmp, stringToAtom("name"), &tmp2);

	if (tmp2 == (Proto) stringToAtom("SlotList")) {
	    addToNode(tmp, newNode);
	} else {
	    privateTree = createNode( (Proto) stringToAtom("SlotList"));
	    addToNode(privateTree, newNode);
	    insertNodeAt(codeObj, privateTree, 0);
	}
    }

    // Make the compiler happy
    return proto;
}


//
//  setupKeywordSlots - Make the parameters to a keyword slot actual members in a slot
//                      list for the object.
// 

Proto setupKeywordSlots (Proto parseTree)
{

    int i,j;
    int isKeywordSlot = 0;
    Proto p1;
    Proto tmp;
    Proto tmp2;
    Proto slotName;
    Proto argName;
    Proto privateTree;
    Proto newNode;
    Proto proto = parseTree;
    Proto codeObj;
    Proto words;
    int   argPosition;
    int   count;
    Proto keyObj;
    char  buff[8];
    
    

    // Determine the type of reference this is
    p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }

    j = nodeCount(proto);

    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if (tmp == (Proto) stringToAtom("Slot")) {
	nodeGet(proto, 0, &tmp);
	if (tmp == (Proto) stringToAtom("Keyword-Slot")) {
	    isKeywordSlot = 1;
	}
    }

    // enumerate through the proto
    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	setupKeywordSlots(tmp);
    }
    
    if (isKeywordSlot) {
	// Get the 'KeywordSlot'
	nodeGet(proto, 1, &keyObj);
	
	// Loop through to build the slot name and the argument names

	words = createNode( (Proto) stringToAtom("wordlist") );
	
	argPosition = 8;
	
	count = nodeCount(keyObj);
	// Don't need the last two items
	count--;
	count--;
	
	// Get the 'Object', which will be used later
	nodeGet(keyObj, (count+2), &codeObj);

	i = 0;
	while (1) {
	    // Get the slotName
	    nodeGet(keyObj, i, &slotName);
	    addToNode(words, slotName);
	    i++;
	    
	    nodeGet(keyObj, i, &tmp2);
	    assert(tmp2 == (Proto) stringToAtom("Identifier"));
	    i++;

	    // Get the argName
	    nodeGet(keyObj, i, &argName);
	    i++;
	
	    newNode = createNode( (Proto) stringToAtom("Slot"));
	    addToNode(newNode, (Proto) stringToAtom("Arg-Slot"));
	    addToNode(newNode, argName);
	    
	    // Also add the position in the stack frame for the argument
	    sprintf(buff, "%d", argPosition);
	    argPosition++;
	    addToNode(newNode, (Proto) stringToAtom(buff));
	    
	    // See if there is already a slot list
	    nodeGet(codeObj, 0, &tmp);
	    objectGetSlot(tmp, stringToAtom("name"), &tmp2);
	    
	    if (tmp2 == (Proto) stringToAtom("SlotList")) {
		addToNode(tmp, newNode);
	    } else {
		privateTree = createNode( (Proto) stringToAtom("SlotList"));
		addToNode(privateTree, newNode);
		insertNodeAt(codeObj, privateTree, 0);
	    }

	    // End the loop
	    if (i > count) {
		break;
	    }
	}

	count = nodeCount(words);
	stringBuffer[0] = 0;
	for (i = 0; i <= count; i++) {
	    nodeGet(words, i, &tmp);
	    strcat(stringBuffer, atomToString((Atom) tmp));
	}
	// Annotate the object
	annotateNode(keyObj, stringToAtom("slotName"), (Proto) stringToAtom(stringBuffer));
    }

    // Make the compiler happy
    return proto;
}


//
//  setupArgSlots - Give the argument slots proper integer positions on the stack
// 

Proto setupArgSlots (Proto parseTree)
{
    int i,j;
    int isSlotList = 0;
    Proto p1;
    Proto tmp;
    Proto tmp2;
    Proto proto = parseTree;
    Proto slot;
    char * p;
    int count;
    char buff[8];
    
    

    // Determine the type of reference this is
    p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }

    j = nodeCount(proto);

    // Tag the arg slots
    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if (tmp == (Proto) stringToAtom("SlotList")) {
	isSlotList = 1;
    }

    // enumerate through the proto
    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	setupArgSlots(tmp);
    }
    
    if (isSlotList) {
	count = 8;
	for (i = 0; i <= j; i++) {
	    nodeGet(proto, i, &slot);
	    objectGetSlot(slot, stringToAtom("name"), &tmp2);
	    if (tmp2 == (Proto) stringToAtom("Slot")) {
		nodeGet(slot, 0, &tmp2);
		if (tmp2 == (Proto) stringToAtom("Arg-Slot")) {
		    nodeGet(slot, 1, &tmp2);
		    p = atomToString((Atom) tmp2);
		    sprintf(buff, "%d", count);
		    if (p[0] == ':') {
			p++;	// Get rid of the colon
			nodeSet(slot, 1, (Proto) stringToAtom(p));
			addToNode(slot,  (Proto) stringToAtom(buff));
		    } else {
			nodeSet(slot, 2, (Proto) stringToAtom(buff));
		    }
		    count++;
		}
	    }
	}
    }
    
    // Make the compiler happy
    return proto;
}



//
//  setupMethods    - Add the slotName and hopefully the locals and param counts
//                    to the Code Object nodes
// 

Proto setupMethods (Proto parseTree)
{

    int i,j;
    int isObjectUnarySlot   = 0;
    int isObjectBinarySlot  = 0;
    int isObjectKeywordSlot = 0;
//    int isBlock            = 0;
    Proto p1;
    Proto tmp;
//    Proto tmp2;
    Proto slotName;
    Proto proto = parseTree;
    Proto codeObj;
    Proto annotate;
    
    int paramCount, localCount;
    

    // Determine the type of reference this is
    p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }


    j = nodeCount(proto);

    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if (tmp == (Proto) stringToAtom("Slot")) {
	nodeGet(proto, 0, &tmp);
	if (tmp == (Proto) stringToAtom("Unary-Slot")) {
	    isObjectUnarySlot = 1;
	}
	if (tmp == (Proto) stringToAtom("Binary-Slot")) {
	    isObjectBinarySlot = 1;
	}
	if (tmp == (Proto) stringToAtom("Keyword-Slot")) {
	    isObjectKeywordSlot = 1;
	}
    }

    // enumerate through the proto
    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	setupMethods(tmp);
    }
    
    if (isObjectUnarySlot) {

	paramCount = 0;
	localCount = 0;
	
	// Get the slot name
	nodeGet(proto, 1, &slotName);
	
	// Get the contained object 
	nodeGet(proto, 2, &tmp);
	
	// Get the UnarySlot (not the Unary-Slot) proto
	// it's '1' slot is the 'Object'
	nodeGet(tmp, 1, &codeObj);
	
	// Count the parameters and locals
	countArgsLocals(codeObj, &paramCount, &localCount);
	// There shouldn't be any parameters
	assert(!paramCount);
	
	
	// Annotate the CodeBlock
	annotateNode(codeObj, stringToAtom("slotName"), slotName);
	annotateNode(codeObj, stringToAtom("localCount"), objectNewInteger(localCount));
	annotateNode(codeObj, stringToAtom("paramCount"), objectNewInteger(paramCount));
    }
    else if (isObjectBinarySlot) {
	paramCount = 1;
	localCount = 0;
	
	// Get the 'Binary-Slot'
	nodeGet(proto, 1, &tmp);
	
	// Get the slotName
	nodeGet(tmp, 0, &slotName);
	
	// Get the 'Object'
	nodeGet(tmp, 3, &codeObj);

	// Count the parameters and locals
	countArgsLocals(codeObj, &paramCount, &localCount);
	// There shouldn't be any parameters defined in the object
	// it is defined in the initializer
	assert((paramCount == 1));
	
	// Annotate the CodeBlock
	annotateNode(codeObj, stringToAtom("slotName"), slotName);
	annotateNode(codeObj, stringToAtom("localCount"), objectNewInteger(localCount));
	annotateNode(codeObj, stringToAtom("paramCount"), objectNewInteger(paramCount));
    }
    else if (isObjectKeywordSlot) {
	paramCount = 0;
	localCount = 0;
	
	// Get the 'KeywordSlot'
	nodeGet(proto, 1, &tmp);
	
	i = nodeCount(tmp);
	assert(objectGetSlot(tmp, stringToAtom("_annotate"), &annotate));
	assert(objectGetSlot(annotate, stringToAtom("slotName"), &slotName));

	nodeGet(tmp, i, &codeObj);
	
	// Count the parameters and locals
	countArgsLocals(codeObj, &paramCount, &localCount);

	// Annotate the CodeBlock
	annotateNode(codeObj, stringToAtom("slotName"), slotName);
	annotateNode(codeObj, stringToAtom("localCount"), objectNewInteger(localCount));
	annotateNode(codeObj, stringToAtom("paramCount"), objectNewInteger(paramCount));
    }
    
    // Make the compiler happy
    return proto;
}

//
//  setupBlocks     - Add the slotName and hopefully the locals and param counts
//                    to the Code Object nodes
// 

Proto setupBlocks (Proto parseTree)
{

    int i,j;
    int isBlock            = 0;
    Proto p1;
    Proto tmp;
    Proto tmp2;
    Proto proto = parseTree;
    Proto objectCount;
    Proto blockObj;
    
    int paramCount, localCount;
    

    // Determine the type of reference this is
    p1 = PROTO_REFVALUE(proto);
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    if (PROTO_REF_ATOM == i) {
	return proto;
    }

    if (PROTO_REF_PROTO != i) {
    	assert(0);
    }

    if (PROTO_ISFORWARDER(proto)) {
        assert(0);
    }

    objectGetSlot(proto, stringToAtom("_count"), &objectCount);
    j = objectIntegerValue(objectCount);

    // A fairly complex way of determining if a codeblock is part of
    // a block or a method
    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if (tmp == (Proto) stringToAtom("Receiver")) {
	nodeGet(proto, 0, &tmp);
	if (tmp == (Proto) stringToAtom("Object")) {
	    nodeGet(proto, 1, &tmp2);
	    i = nodeCount(tmp2);
	    nodeGet(tmp2, i, &tmp);
	    objectGetSlot(tmp, stringToAtom("name"), &tmp2);
	    if (tmp2 == (Proto) stringToAtom("CodeBlock")) {
		isBlock = 1;
	    }
	}
    }
    
		    
    // enumerate through the proto
    for (i = 0; i <= j; i++) {
	nodeGet(proto, i, &tmp);
	setupBlocks(tmp);
    }
    
    if (isBlock) {
	
	paramCount = 0;
	localCount = 0;
	
	// Get the Object node
	nodeGet(proto, 1, &blockObj);

	// Count the parameters and locals
	countArgsLocals(blockObj, &paramCount, &localCount);
	
	// Annotate the CodeBlock
	annotateNode(blockObj, stringToAtom("slotName"), (Proto) stringToAtom("**Block"));
	annotateNode(blockObj, stringToAtom("localCount"), objectNewInteger(localCount));
	annotateNode(blockObj, stringToAtom("paramCount"), objectNewInteger(paramCount));
    }

    
    // Make the compiler happy
    return proto;
}


// This function will count the number of slots in an object in order
// to determine the number of local variables a method or block is using.
void countArgsLocals (Proto obj, int * argCount, int * localCount )
{
    Proto tmp;
    Proto tmp2;
    Proto tmp3;
    int i,j;
    
    *argCount = 0;
    *localCount = 0;
    

    // The first slot is always the slot or slotList
    nodeGet(obj, 0, &tmp);
		
    // Check to see if it is a SlotList
    objectGetSlot(tmp, stringToAtom("name"), &tmp2);
    if (tmp2 == (Proto) stringToAtom("SlotList")) {
	j = nodeCount(tmp);

	// enumerate through the proto
	for (i = 0; i <= j; i++) {
	    nodeGet(tmp, i, &tmp2);
	    nodeGet(tmp2, 0, &tmp3);
	    if (tmp3 == (Proto) stringToAtom("Arg-Slot")) {
		(*argCount)++;
	    } else {
		(*localCount)++;
	    }
	}
    }
}




//
//  codeGen - generate a final soup for this tree, this could get to be a slightly
//            big function
//            NOTE - even though this is called codeGen - that is only part of it's
//            functionality because a soup is composed of objects and code. I'm just
//            emphasizing the more difficult side of the function... the code generation
// 

Proto codeGen (Proto parseTree, int addStart)
{

    Proto tmp;


    Proto soup;
    Proto start;
    Proto root;
    
    Proto startCode;
    

    // Build the soup
    soup = realCodeGen(parseTree);

    // assert(0);
    
    if (addStart) {
	
	// Finish the soup with a little spice (i.e. this a slot that the vm 
	// will look for to start the program running.)
	start = objectNew(0);
	objectGetSlot(soup, stringToAtom("Root"), &root);
	objectSetSlot(root, stringToAtom("_Start"), PROTO_ACTION_GET, (uint32) start);
	
	objectSetSlot(start, stringToAtom("Stack"), PROTO_ACTION_GET, (uint32) objectNewInteger(2048));
	
	
	startCode = createNode( (Proto) stringToAtom("Ignore This Slot"));
	objectSetSlot(start, stringToAtom("_assemble"), PROTO_ACTION_GET, (uint32) startCode);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("slotName"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) stringToAtom("_Start"));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("setupFrame"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) objectNewInteger(0));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("nop"));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) stringToAtom("Start"));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushSelf"));
	addToNode(startCode, tmp);
	
	// Push the arg count
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) objectNewInteger(0));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("call"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("activateSlot"));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pop"));
	addToNode(startCode, tmp);
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("break"));
	addToNode(startCode, tmp);
    }
    
    return soup;
}

//
//  realCodeGen - generate a final soup for this tree, this could get to be a slightly
//            big function
// 

Proto realCodeGen(Proto parseTree)
{

    Proto tmpObj;
    Proto nodeName;
    int i;
    Proto proto = parseTree;

    

    // Begin the process of code generation...
        
    if (proto == nil) {
	printString(" ERROR a <nil> ");
	assert(0);
    }

    i = objectGetSlot(proto, stringToAtom("name"), &nodeName);
    assert(i);

    if ((Proto) stringToAtom("Soup") == nodeName) {
	nodeGet(proto, 0, &tmpObj);
	return cgBuildObject(tmpObj);
    }

    assert(0);
}


Proto cgBuildObject (Proto parseTree)
{

    int   i,j;
    Proto newObj;
    Proto node;
    Proto nodeName;
    Proto tmpObj;
    Proto annotation;
    int count;
    
    
    Proto proto = parseTree;

    
    newObj = objectNew(0);
    
    i = objectGetSlot(proto, stringToAtom("name"), &nodeName);
    assert(i);
    assert((Proto) stringToAtom("Object") == nodeName);

    // READ THIS
    // The situation The Proto might have two slots: 0 - SlotList, 1 - CodeBlock
    // or it might have one slot: 0 - CodeBlock
    // The following code will handle either

    // Get the slot list and other optional parts
    count = nodeCount(proto);
    i = nodeGet(proto, 0, &node);
    assert(i);
    i = objectGetSlot(node, stringToAtom("name"), &tmpObj);
    assert(i);

    if ((Proto) stringToAtom("SlotList") == tmpObj) {
	j = nodeCount(node);
	
	// enumerate through the proto building it's slots
	for (i = 0; i <= j; i++) {
	    nodeGet(node, i, &tmpObj);
	    cgBuildSlot(tmpObj, newObj);
	}
	
	// Prep the next If statement
	if (0 == count) {
	    // There was only one slot
	    return newObj;
	}
	
	i = nodeGet(proto, 1, &node);
	assert(i);
	i = objectGetSlot(node, stringToAtom("name"), &tmpObj);
	assert(i);
    }

    // If an object contains a codeblock, then the object as a whole
    // is a code object. The cgBuildCodeBlock will handle either a method
    // or a block (since they are essentially the same)
    if ((Proto) stringToAtom("CodeBlock") == tmpObj) {
	assert(objectGetSlot(proto, stringToAtom("_annotate"), &annotation));
	tmpObj = cgBuildCodeBlock(node, annotation);
	objectSetSlot(newObj, (Atom) stringToAtom("_assemble"), PROTO_ACTION_GET, (uint32)tmpObj);
    }

    return newObj;
}


void cgBuildSlot (Proto parseTree, Proto newObj)
{

    int i;
    Proto tmp;
    Proto tmp2;
    Proto newValue;
    Proto slotName;
    Proto annotate;
    
    
    Proto proto = parseTree;

    
    nodeGet(proto, 0, &tmp);
    
    if ( (Proto) stringToAtom("Identifier-Slot") == tmp) {
	nodeGet(proto, 1, &tmp);
	objectSetSlot(newObj, (Atom) tmp, PROTO_ACTION_GET, (uint32) nil);
	return;
    }
    
    if ( (Proto) stringToAtom("Data-Get-Slot") == tmp) {
	nodeGet(proto, 1, &tmp);
	nodeGet(proto, 2, &tmp2);
	// The node is receiver
	newValue = cgBuildReceiver(tmp2);
	objectSetSlot(newObj, (Atom) tmp, PROTO_ACTION_GET, (uint32) newValue);
	return;
    }
    if ( (Proto) stringToAtom("Data-SetGet-Slot") == tmp) {
	nodeGet(proto, 1, &tmp);
	nodeGet(proto, 2, &tmp2);
	// The node is receiver
	newValue = cgBuildReceiver(tmp2);
	objectEasySetSlot(newObj, (Atom) tmp, (uint32) newValue);
	return;
    }
    if ( (Proto) stringToAtom("Arg-Slot") == tmp) {
	nodeGet(proto, 1, &slotName);
	nodeGet(proto, 2, &tmp2);
	// Get the order in the frame
	i = atoi(atomToString((Atom) tmp2));
	tmp2 = objectNewInteger(i);
	objectSetSlot(newObj, (Atom) slotName, PROTO_ACTION_GETFRAME, (uint32) tmp2);
	return;
    }

    if ( (Proto) stringToAtom("Unary-Slot") == tmp) {
	nodeGet(proto, 1, &slotName);
	// Get the 'node' holding the object
	nodeGet(proto, 2, &tmp2);
	// One more level down the 'node' holding the object
	nodeGet(tmp2, 1, &tmp2);
	newValue = cgBuildObject(tmp2);
	objectSetSlot(newObj, (Atom) slotName, PROTO_ACTION_GET, (uint32) newValue);
	return;
    }
    if ( (Proto) stringToAtom("Binary-Slot") == tmp) {
	// Get the 'node' holding the object
	nodeGet(proto, 1, &tmp2);
	nodeGet(tmp2, 0, &slotName);
	nodeGet(tmp2, 3, &tmp);
	newValue = cgBuildObject(tmp);
	objectSetSlot(newObj, (Atom) slotName, PROTO_ACTION_GET, (uint32) newValue);
	return;
    }
    
    if ( (Proto) stringToAtom("Keyword-Slot") == tmp) {
	// Get the 'node' holding the object
	nodeGet(proto, 1, &tmp2);
	i = nodeCount(tmp2);
	assert(objectGetSlot(tmp2, stringToAtom("_annotate"), &annotate));
	assert(objectGetSlot(annotate, stringToAtom("slotName"), &slotName));
	nodeGet(tmp2, i, &tmp);
	newValue = cgBuildObject(tmp);
	objectSetSlot(newObj, (Atom) slotName, PROTO_ACTION_GET, (uint32) newValue);
	return;
    }

    // We shouldn't get here
    assert(0);
}


Proto cgBuildReceiver (Proto parseTree)
{

    Proto tmp;
    int i;

    Proto proto = parseTree;

    
    nodeGet(proto, 0, &tmp);
    
    if ( (Proto) stringToAtom("String") == tmp) {
	nodeGet(proto, 1, &tmp);
	return tmp;
    }
    if ( (Proto) stringToAtom("Object") == tmp) {
	nodeGet(proto, 1, &tmp);
	return cgBuildObject(tmp);
    }
    if ( (Proto) stringToAtom("Number") == tmp) {
	nodeGet(proto, 1, &tmp);
	i = atoi(atomToString((Atom) tmp));
	return objectNewInteger(i);
    }
    if ( (Proto) stringToAtom("Nil Expression") == tmp) {
    	return nil;
    }
    if ( (Proto) stringToAtom("Expression") == tmp) {
	nodeGet(proto, 1, &tmp);
	return cgBuildExpression(tmp);
    }

    // We shouldn't get here
    assert(0);
}

Proto cgBuildReceiverAssembly (Proto parseTree)
{

    int i;
    Proto tmp;
    Proto object;
    Proto assemble;
    Proto newObj;
    
    Proto proto = parseTree;

    
    objectGetSlot(proto, stringToAtom("name"), &tmp);
    if ( (Proto) stringToAtom("Receiver") == tmp) {
	object = cgBuildReceiver(proto);
    }
    if ( (Proto) stringToAtom("Binary") == tmp) {
	object = cgBuildBinary(proto);
    }
    if ( (Proto) stringToAtom("Unary") == tmp) {
	object = cgBuildUnary(proto);
    }

    i  = PROTO_REFTYPE(object);

    if (PROTO_REF_PROTO == i) {
	if (object != nil) {
	    objectGetSlot(object, stringToAtom("name"), &tmp);
	    if ( (Proto) stringToAtom("__assemble") == tmp) {
		return object;
	    }
	}
	
    }
	
    assemble = createNode( (Proto) stringToAtom("__assemble"));
    newObj = objectNew(0);
    
    if (object == nil) {
	objectSetSlot(newObj, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushSelf"));
	addToNode(assemble, newObj);
	return assemble;
    }
    
    objectSetSlot(newObj, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
    objectSetSlot(newObj, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
    objectSetSlot(newObj, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) object);
    addToNode(assemble, newObj);
    return assemble;
}



Proto cgBuildCodeBlock (Proto parseTree, Proto annotation)
{
    int i,j;
    Proto tmp;
    Proto tmp2;
    Proto assemble;
    Proto assemble2;
    Proto slotName;
    Proto localCount;
    Proto paramCount;
    
        
    Proto proto = parseTree;
    
    
    assemble = createNode( (Proto) stringToAtom("__assemble"));

    i = objectGetSlot(annotation, stringToAtom("localCount"), &localCount);
    assert(i);
    
    i = objectGetSlot(annotation, stringToAtom("paramCount"), &paramCount);
    assert(i);
    
    i = objectGetSlot(annotation, stringToAtom("slotName"), &slotName);
    assert(i);
    
    

    // Slot Name
    tmp = objectNew(0);
    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("slotName"));
    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) slotName);
    addToNode(assemble, tmp);
    
    // Setup Frame for Locals
    tmp = objectNew(0);
    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("setupFrame"));
    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) localCount);
    addToNode(assemble, tmp);
    
    //// We will need to initialize those locals at one point...
    //// For now they are all nil
    //// Will need to initialize those locals, at one point....


    j = nodeCount(proto);
    
    // enumerate through the code block - an array of expressions
    i = 0;
    while (1) {
	nodeGet(proto, i, &tmp2);
	assemble2 = cgBuildExpression(tmp2);

	appendNodeToNode(assemble2, assemble);
	
	i++;
	

	if (i > j) {
	    tmp = objectNew(0);
	    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("restoreFrame"));
	    addToNode(assemble, tmp);
	    
	    tmp = objectNew(0);
	    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("return"));
	    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) objectNewInteger(2));
	    addToNode(assemble, tmp);
	    break;
	} else {
	    // Each expression's result is not used
	    tmp = objectNew(0);
	    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pop"));
	    addToNode(assemble, tmp);
	}
    }
    
    return assemble;
}

Proto cgBuildExpression (Proto parseTree)
{
    int i;
    Proto tmp;
    Proto assemble;
    Proto node;
    
        
    Proto proto = parseTree;
    
    
    i = nodeGet(proto, 0, &node);
    assert(i);

    i = objectGetSlot(node, stringToAtom("name"), &tmp);
    assert(i);

    if ( (Proto) stringToAtom("Unary") == tmp) {
	assemble = cgBuildUnary(node);
	return assemble;
    }
    
    if ( (Proto) stringToAtom("Binary") == tmp) {
	assemble = cgBuildBinary(node);
	return assemble;
    }

    if ( (Proto) stringToAtom("Message") == tmp) {
	assemble = cgBuildMessage(node);
	return assemble;
    }
    
    // Shouldn't get here
    assert(0);
}




Proto cgBuildUnary (Proto parseTree)
{
    int i,j;
    Proto tmp;
    Proto operand;
    Proto target;
    Proto assemble;
    
    Proto proto = parseTree;
    
    j = 0;
    i = nodeCount(proto);

    assemble = createNode( (Proto) stringToAtom("__assemble"));
    
    while (1) {
	
	if (0 == j) {
	    // 0 - get the first object handled by code gen
	    nodeGet(proto, 0, &tmp);
	    target = cgBuildReceiverAssembly(tmp);
	    j++;
	    j++;
	    
	    nodeGet(proto, 1, &operand);
	    
	    tmp = objectNew(0);
	    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	    objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) operand);
	    addToNode(assemble, tmp);
	    appendNodeToNode(target, assemble);
	} else {
	    // Access the array by the keys in order
	    nodeGet(proto, j, &operand);
	    
	    tmp = objectNew(0);
	    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	    objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) operand);
	    addToNode(assemble, tmp);
	    
	    tmp = objectNew(0);
	    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("swap"));
	    addToNode(assemble, tmp);
	    j++;
	}
	
	// Push the argument count
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) objectNewInteger(2));
	addToNode(assemble, tmp);
	
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("call"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("activateSlot"));
	addToNode(assemble, tmp);
	
	if (j > i) {
	    break;
	}
    }
    
    return assemble;
}






Proto cgBuildBinary (Proto parseTree)
{
    int i,j;
    Proto tmp;
    Proto operand;
    Proto operator;
    Proto target;
    Proto assemble;
    
    Proto proto = parseTree;
    
    i = nodeCount(proto);
    j = 0;
    
    assemble = createNode( (Proto) stringToAtom("__assemble"));
    
    // Need to do leftToRight, push right, push left, push message

    // Work backwards
    // An expression like 3 + 4 - 1 is (3 + 4) - 1
    // However since this is a stack machine
    // the code on the stack will be   1-4+3CC  
    // (C represents Call), otherwise there would be lots of swaps on
    // the stack
    // Problem: (3 + 4) - 5 - 1
    while (1) {
	
	if (j == 0) {
	    nodeGet(proto, j, &tmp);
	    target = cgBuildReceiverAssembly(tmp);
	    j++;
	    // Push the target
	    appendNodeToNode(target, assemble);
	}
	
	
	nodeGet(proto, j, &operator);
	j++;

	nodeGet(proto, j, &tmp);
	operand = cgBuildReceiverAssembly(tmp);
	j++;
	// Push the operand
	appendNodeToNode(operand, assemble);
	// swap
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("swap"));
	addToNode(assemble, tmp);
	

	// Push the mesg/operator
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) operator);
	addToNode(assemble, tmp);
	// swap
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("swap"));
	addToNode(assemble, tmp);

	
	// Push the argument count
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
	objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) objectNewInteger(2));
	addToNode(assemble, tmp);
	
	
	tmp = objectNew(0);
	objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("call"));
	objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("activateSlot"));
	addToNode(assemble, tmp);
	
	if (j > i) {
	    break;
	}
    }
    
    return assemble;

}



Proto cgBuildMessage(Proto parseTree)
{
    int i,j;
    Proto tmp;
    Proto messageName;
    Proto operand;
    Proto target;
    Proto assemble;
    Proto operator;
    
    
    Proto proto = parseTree;
    
    
    // Get all of the arguments prepared
    // Then setup the call to the method
    
    // Just like Binary, we work backwards since that is how
    // arguments are pushed on the stack
    
    // Essentially, go through each operand in reverse and add
    // them as operands
    // Then also push the method name (after it has been 
    // concatenated)
    // push self (or target object)
    // make call
    
    i = nodeCount(proto);
    j = i;
    
    
    assemble = createNode( (Proto) stringToAtom("__assemble"));
    messageName = createNode( (Proto) stringToAtom("MessageName"));
    
    // Go in reverse and push the arguments while collecting the 
    // Messages
    while (1) {
	
	nodeGet(proto, j, &tmp);
	operand = cgBuildReceiverAssembly(tmp);
	j--;
	appendNodeToNode(operand, assemble);
	
	nodeGet(proto, j, &operator);
	addToNode(messageName, operator);
	j--;
	
	if (0 == j) {
	    break;
	}
    }
    
    
    // Ok it is now time to generate the message
    
    // Go through the messageName forward and write that string
    // as the selector
    // Go through the args in reverse and push the 
    i = nodeCount(messageName);
    
    j = i;
    stringBuffer[0] = 0;
    
    while (1) {
	nodeGet(messageName, j, &operand);
	
	strcat(stringBuffer, atomToString((Atom) operand));
	
	j--;
	
	if (j < 0) {
	    break;
	}
    }
    
    tmp = objectNew(0);
    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
    objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) stringToAtom(stringBuffer));
    addToNode(assemble, tmp);
    
    // Finally, push the target
    nodeGet(proto, 0, &tmp);
    target = cgBuildReceiverAssembly(tmp);
    appendNodeToNode(target, assemble);

    // Push the argument count
    tmp = objectNew(0);
    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("pushimmediate"));
    objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("object"));
    objectSetSlot(tmp, stringToAtom("operand"), PROTO_ACTION_GET, (uint32) objectNewInteger(i+2));
    addToNode(assemble, tmp);
    
	
    tmp = objectNew(0);
    objectSetSlot(tmp, stringToAtom("op"), PROTO_ACTION_GET, (uint32) stringToAtom("call"));
    objectSetSlot(tmp, stringToAtom("type"), PROTO_ACTION_GET, (uint32) stringToAtom("activateSlot"));
    addToNode(assemble, tmp);
    
    return assemble;
}

