//
// svm32.h
//
// This is the description of the functions for the Stack-VM 32 bit
// 
//
// $Id: svm32.h,v 1.6 2000/12/30 10:39:46 dru Exp $
//
// $Log: svm32.h,v $
// Revision 1.6  2000/12/30 10:39:46  dru
// This is the checkin that has working blocks!
// Plus tons of other fixes!
//
// Revision 1.5  2000/08/27 19:49:28  dru
// This is the result of a major change in the code to handle
// the new reference size, and the changes in the way the
// code handles activations.
//
// Revision 1.4  1999/09/07 08:48:47  dru
// Lots of changes to get the debugger working
//
// Revision 1.3  1999/08/17 15:43:01  dru
// Checking stuff in - not ready for GTK debugger
//
// Revision 1.2  1999/04/28 06:28:28  dnelson
// Changes for doing argument count checking
//
// Revision 1.1  1999/03/04 05:54:48  dnelson
// This is the first checkin of some code that is pretty stable
// This is the stack based VM for Bubl/Aqua
//
//
//
//
//


#ifndef H_BUBLSVM32
#define H_BUBLSVM32


#define CC_CARRY 0x01
#define CC_ZERO  0x02
#define CC_SIGN  0x04
#define CC_OF	 0x08

struct _vm32cpu {
	unsigned int pc;	/* Program Counter */
	unsigned int cc;	/* Condition Code */
	unsigned int sp;	/* Stack Pointer */
	unsigned int fp;	/* Frame Pointer */
	unsigned int haltflag;	/* Halt Flag :-) */
	// The following two are for branches
	unsigned int trueObj;	/* Value that references true obj */
	unsigned int falseObj;	/* Value that references false obj */
	void (*breakHook) (struct _vm32cpu *);	/* Break Handler */
	void (*invalidHook)(struct _vm32cpu *);	/* Invalid Instruction Handler */
	void (*argCountHook)(struct _vm32cpu *);	/* Invalid message argument count hook */
	unsigned int (*pushHook)(struct _vm32cpu *, unsigned int);	/* Hook for special pushImmediate handling */
};


typedef struct _vm32cpu VM32Cpu;


inline void push(VM32Cpu * cpu, unsigned int v);
inline unsigned int pull(VM32Cpu * cpu);
inline unsigned int stackAt(VM32Cpu * cpu, unsigned int i);

inline unsigned int getFrameAt(VM32Cpu * cpu, int i);
inline unsigned int getFPRelativeAt(unsigned int fp, int i);

inline unsigned int setFrameAt(VM32Cpu * cpu, int pos, unsigned int value);
inline unsigned int setFPRelativeAt(unsigned int fp, int pos, unsigned int value);


void init (VM32Cpu * cpu);

void reset (VM32Cpu * );
int run (VM32Cpu * );
int singleStep (VM32Cpu * );
int halt (VM32Cpu *);


// User level routines 
inline void VMReturn(VM32Cpu * cpu, unsigned int returnValue, unsigned int moveCount);

// Debugger Support
int disassembleCpu(VM32Cpu * cpu, char * * opcode,  int * length);


#endif
